/* GTK - The GIMP Toolkit
 * Copyright (C) 1995-1997 Peter Mattis, Spencer Kimball and Josh MacDonald
 * GtkColorButton Copyright (C) 2000 Stefan Ondrejicka <ondrej@idata.sk>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "gtkclrbutton.h"
#include <gtk/gtk.h>
#include <string.h>


#define RED     0
#define GREEN   1
#define BLUE    2
#define OPACITY 3


enum {
  COLOR_CHANGED,
  LAST_SIGNAL
};


static void gtk_color_button_class_init		(GtkColorButtonClass	*klass);
static void gtk_color_button_init		(GtkColorButton		*cbutton);
static void gtk_color_button_destroy		(GtkObject		*object);
static void gtk_color_button_color_changed	(GtkColorButton		*cbutton);
static void gtk_color_button_preview_paint	(GtkColorButton		*cbutton);
static void gtk_color_button_preview_resize	(GtkWidget		*w,
						 GtkAllocation		*allocation,
						 GtkColorButton		*cbutton);
static void gtk_color_button_select_color	(GtkColorButton		*cbutton);
static void gtk_color_button_colorsel_hide	(GtkObject		*object,
						 GtkColorButton         *cbutton);
static void gtk_color_button_colorsel_ok	(GtkObject		*object,
						 GtkColorButton         *cbutton);

static gint color_button_signals[LAST_SIGNAL] = { 0 };
static GtkButtonClass *parent_class = NULL;

GtkType gtk_color_button_get_type (void)
{

	static GtkType color_button_type = 0;

	if (!color_button_type)
	{
		static const GTypeInfo color_button_info =
		{
			sizeof (GtkColorButtonClass),
			NULL, /* base_init */
			NULL, /* base_finalize */
			(GClassInitFunc) gtk_color_button_class_init,
			NULL, /* class_finalize */
			NULL, /* class_data */
			sizeof(GtkColorButton),
			0,    /* n_preallocs */
			(GInstanceInitFunc) gtk_color_button_init,
		};

		color_button_type = g_type_register_static 
						(GTK_TYPE_BUTTON, "GtkColorButton",&color_button_info,0);
        }

        return color_button_type;;
}

static void gtk_color_button_class_init (GtkColorButtonClass *klass)
{
	GtkObjectClass *object_class;

	object_class = (GtkObjectClass*) klass;
	parent_class = gtk_type_class (GTK_TYPE_BUTTON);
	object_class->destroy = gtk_color_button_destroy;

	color_button_signals[COLOR_CHANGED] = 
		g_signal_new ("color_changed",
				G_TYPE_FROM_CLASS (klass),
				G_SIGNAL_RUN_FIRST, 
				G_STRUCT_OFFSET (GtkColorButtonClass, color_changed),
				NULL, NULL,
				g_cclosure_marshal_VOID__POINTER,
				G_TYPE_NONE, 1, G_TYPE_POINTER);

// Obselete
// gtk_object_class_add_signals (object_class, color_button_signals, LAST_SIGNAL);

	klass->color_changed = NULL;
}

static void gtk_color_button_init (GtkColorButton *cbutton)
{
	GTK_WIDGET_SET_FLAGS (cbutton, GTK_CAN_FOCUS | GTK_RECEIVES_DEFAULT);
	// Causes the app to die if uncommented.
//	GTK_WIDGET_UNSET_FLAGS (cbutton, GTK_NO_WINDOW);
	cbutton->color.pixel = 0;
	cbutton->color.red = 0;
	cbutton->color.green = 0;
	cbutton->color.blue = 0;

	cbutton->title = NULL;
	cbutton->colorsel = NULL;
	cbutton->had_colorsel = FALSE;

	cbutton->draw = gtk_drawing_area_new();
	gtk_drawing_area_size(GTK_DRAWING_AREA(cbutton->draw),30,20);
	gtk_container_add (GTK_CONTAINER (cbutton) , cbutton->draw);
	gtk_widget_show(cbutton->draw); 
	
	gtk_signal_connect_after (GTK_OBJECT (cbutton), "clicked",
		GTK_SIGNAL_FUNC (gtk_color_button_select_color), NULL); 
}

GtkWidget *gtk_color_button_new (void)
{
	return GTK_WIDGET (g_object_new (GTK_TYPE_COLOR_BUTTON, NULL));
}

void gtk_color_button_set_title (GtkColorButton *cbutton, gchar *title)
{
	g_return_if_fail (cbutton != NULL);
	g_return_if_fail (GTK_IS_COLOR_BUTTON (cbutton));

	g_free(cbutton->title);
	cbutton->title = g_strdup(title);
}

void gtk_color_button_set_color (GtkColorButton *cbutton, GdkColor *color)
{
	g_return_if_fail (cbutton != NULL);
	g_return_if_fail (GTK_IS_COLOR_BUTTON (cbutton));
	g_return_if_fail (color != NULL);

	memcpy(&cbutton->color,  color, sizeof(GdkColor));
	gtk_color_button_preview_paint (cbutton);
	gtk_color_button_color_changed (cbutton);
}

void gtk_color_button_get_color (GtkColorButton *cbutton, GdkColor *color)
{
	g_return_if_fail (cbutton != NULL);
	g_return_if_fail (GTK_IS_COLOR_BUTTON (cbutton));
	g_return_if_fail (color != NULL);

	memcpy(color, &cbutton->color, sizeof(GdkColor));
}

static void gtk_color_button_destroy (GtkObject *object)
{
	g_free(GTK_COLOR_BUTTON(object)->title);

	if (GTK_COLOR_BUTTON(object)->colorsel)
		gtk_widget_destroy(GTK_COLOR_BUTTON(object)->colorsel);
}

static void gtk_color_button_color_changed (GtkColorButton *cbutton)
{
	gtk_signal_emit (GTK_OBJECT (cbutton), color_button_signals[COLOR_CHANGED], &cbutton->color);
}

static void gtk_color_button_preview_paint (GtkColorButton *cbutton)
{
	// This probably isn't the best way to do this, but 
	// a) it works,
	// b) it's simple, and
	// c) it's not deprecated. Previously, a GtkPreview was being used and I'd like
	// 		as much of ChBg to compile under GTK_DISABLE_DEPRECATED as possible.
	// If there's a good reason it shouldn't be done this way (and I'm sure there is)
	// could you fix it, and/or let me know what the better way is?
				
	gtk_widget_modify_bg(cbutton->draw,GTK_STATE_NORMAL,&(cbutton->color));
	
}

static void gtk_color_button_preview_resize (GtkWidget *w, GtkAllocation *allocation, GtkColorButton *cbutton)
{
	gtk_color_button_preview_paint (cbutton);
}

static void gtk_color_button_colorsel_hide (GtkObject *object, GtkColorButton *cbutton)
{
	if (cbutton->colorsel)
		gtk_widget_destroy(cbutton->colorsel);
}

static void gtk_color_button_colorsel_ok (GtkObject *object, GtkColorButton *cbutton)
{
	gdouble color[4];

	gtk_color_selection_get_color(GTK_COLOR_SELECTION(
		GTK_COLOR_SELECTION_DIALOG(cbutton->colorsel)->colorsel) , color);

	cbutton->color.red = color[RED] * 65535;
	cbutton->color.green = color[GREEN] * 65535;
	cbutton->color.blue = color[BLUE] * 65535;

	gtk_color_button_preview_paint (cbutton);
	gtk_color_button_color_changed (cbutton);

	gtk_widget_destroy(cbutton->colorsel);
}

static void gtk_color_button_select_color (GtkColorButton *cbutton)
{
	if (!cbutton->colorsel)
	{
		gdouble color[4];

		cbutton->colorsel = gtk_color_selection_dialog_new(cbutton->title);
		gtk_signal_connect (GTK_OBJECT (cbutton->colorsel), "destroy",
			GTK_SIGNAL_FUNC(gtk_widget_destroyed), &cbutton->colorsel);

		if (!cbutton->had_colorsel)
		{
			gtk_signal_connect(GTK_OBJECT(gtk_widget_get_toplevel(GTK_WIDGET(cbutton))),
				"hide", GTK_SIGNAL_FUNC(gtk_color_button_colorsel_hide),
				(gpointer) cbutton);

			cbutton->had_colorsel = TRUE;
		}

		color[RED] = ((gdouble)cbutton->color.red) / 65535.0;
		color[GREEN] = ((gdouble)cbutton->color.green) / 65535.0;
		color[BLUE] = ((gdouble)cbutton->color.blue) / 65535.0;
		color[OPACITY] = 1.0;

		gtk_color_selection_set_color(GTK_COLOR_SELECTION (
			GTK_COLOR_SELECTION_DIALOG (cbutton->colorsel)->colorsel) , color);

		gtk_signal_connect (GTK_OBJECT (GTK_COLOR_SELECTION_DIALOG (cbutton->colorsel)->ok_button),
			"clicked", GTK_SIGNAL_FUNC(gtk_color_button_colorsel_ok), cbutton);

		gtk_signal_connect (GTK_OBJECT (GTK_COLOR_SELECTION_DIALOG (cbutton->colorsel)->cancel_button),
			"clicked", GTK_SIGNAL_FUNC(gtk_color_button_colorsel_hide), cbutton);

	}

	gtk_widget_show (cbutton->colorsel);
	if (GTK_WIDGET_REALIZED (cbutton->colorsel))
		gdk_window_raise (cbutton->colorsel->window);
}

